/*----------------------------------------------------------------------*
 * Bounds Checking for GCC.						*
 * Copyright (C) 1995 Richard W.M. Jones <rjones@orchestream.com>.	*
 *----------------------------------------------------------------------*
 * This program is free software; you can redistribute it and/or modify	*
 * it under the terms of the GNU General Public License as published by	*
 * the Free Software Foundation; either version 2 of the License, or	*
 * (at your option) any later version.					*
 *									*
 * This program is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
 * GNU General Public License for more details.				*
 *									*
 * You should have received a copy of the GNU General Public License	*
 * along with this program; if not, write to the Free Software		*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*
 *----------------------------------------------------------------------*
 * File:
 *	run-includes/unchecked.h
 * Summary:
 *	Macros to allow per-pointer and per-use unchecked pointers.
 * Other notes:
 *	
 * Author      	Date		Notes
 * RWMJ		23/1/95		Initial implementation.
 * RWMJ         17/6/95         Added miscellaneous features.
 *----------------------------------------------------------------------*/

#ifndef _bounds_unchecked_h
#define _bounds_unchecked_h

#if defined(__GNUC__) && defined(__BOUNDS_CHECKING_ON)

/* Make pointer operations unchecked on each use.  */
#define UNCHECKED_PTR_DIFF(p,q)		BOUNDS_CHECKING_OFF_IN_EXPR((p) - (q))
#define UNCHECKED_PTR_PLUS_INT(p,i)	BOUNDS_CHECKING_OFF_IN_EXPR((p) + (i))
#define UNCHECKED_PTR_MINUS_INT(p,i)	BOUNDS_CHECKING_OFF_IN_EXPR((p) - (i))

/* How to switch checking off over a region of the program. You will need
 * 'DEBUG_FEATURES' set in the library. You do:
 *     ..
 *    BOUNDS_CHECKING_OFF;
 *     .. strange pointer operations here ..
 *    BOUNDS_CHECKING_ON;
 */
#define BOUNDS_CHECKING_OFF \
_Pragma("bounds_checking off")

#define BOUNDS_CHECKING_ON  \
_Pragma("bounds_checking on")

/* Switch bounds checking off in a single statement. You do:
 *    BOUNDS_CHECKING_OFF_DURING (stmt);
 * This works for most 'ordinary' statements, ie. not returns, or gotos, etc.
 * For expressions, use:
 *    BOUNDS_CHECKING_OFF_IN_EXPR (expr)
 */
#define BOUNDS_CHECKING_OFF_DURING(stmt) \
  BOUNDS_CHECKING_OFF \
  do { stmt; } while (0) \
  BOUNDS_CHECKING_ON

#define BOUNDS_CHECKING_OFF_IN_EXPR(expr) \
  BOUNDS_CHECKING_OFF \
  (expr) \
  BOUNDS_CHECKING_ON


#if defined(__OPTIMIZE__) && defined(__BOUNDS_INLINE__)

#ifndef __BOUNDS_DEBUG_FEATURES
#define	__BOUNDS_DEBUG_FEATURES	0
#endif
#ifndef __BOUNDS_COLLECT_STATS
#define	__BOUNDS_COLLECT_STATS	0
#endif

#if __BOUNDS_DEBUG_FEATURES
extern int __bounds_debug_print_calls;
#endif

#if __BOUNDS_COLLECT_STATS
extern unsigned __bounds_stats_ptr_plus_int;
extern unsigned __bounds_stats_array_reference;
extern unsigned __bounds_stats_ptr_diff;
extern unsigned __bounds_stats_reference;
extern unsigned __bounds_stats_component_reference;
extern unsigned __bounds_stats_ptr_le_ptr;
extern unsigned __bounds_stats_ptr_lt_ptr;
extern unsigned __bounds_stats_ptr_ge_ptr;
extern unsigned __bounds_stats_ptr_gt_ptr;
extern unsigned __bounds_stats_ptr_ne_ptr;
extern unsigned __bounds_stats_ptr_eq_ptr;
extern unsigned __bounds_stats_ptr_postinc;
extern unsigned __bounds_stats_ptr_preinc;
extern unsigned __bounds_stats_ptr_postdec;
extern unsigned __bounds_stats_ptr_predec;
extern unsigned __bounds_stats_ptr_true;
extern unsigned __bounds_stats_ptr_false;
#endif

#pragma bounds_checking off

#define __BOUNDS_NULL    (void *)0
#define __BOUNDS_ILLEGAL (void *)-2

typedef __SIZE_TYPE__ __bounds_size_t;

typedef struct
{
  void *__bounds_base;		/* Points to base of the object.  */
  void *__bounds_extent;	/* Points to last byte in object+1.  */
  __bounds_size_t __bounds_size;	/* Size of the object.  */
} __bounds_object;

typedef union
{
  void *__bounds_void_pointer;
  char *__bounds_char_pointer;
} __bounds_union_type;

extern __bounds_object *__bounds_find_object (void *__bounds_pointer);

#pragma redefine_extname __dummy_bounds_check_ptr_plus_int \
				__bounds_check_ptr_plus_int

extern void *__dummy_bounds_check_ptr_plus_int (__bounds_object *
						__bounds_obj,
						void *__bounds_pointer,
						int __bounds_offset,
						__bounds_size_t __bounds_size,
						const char *__bounds_filename,
						int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_plus_int (__bounds_object * __bounds_obj,
			     void *__bounds_pointer, int __bounds_offset,
			     __bounds_size_t __bounds_size,
			     const char *__bounds_filename, int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pnew;

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer +=
    (int) __bounds_size *__bounds_offset;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_plus_int (__bounds_obj, __bounds_pointer,
					 __bounds_offset, __bounds_size,
					 __bounds_filename, __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_ptr_plus_int;
#endif
  return __bounds_pnew;
}

extern __inline__ void *
__bounds_check_ptr_plus_int_obj (void *__bounds_pointer,
				 int __bounds_offset,
				 __bounds_size_t __bounds_size,
				 const char *__bounds_filename,
				 int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pnew;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pointer);

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer +=
    (int) __bounds_size *__bounds_offset;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_plus_int (__bounds_obj, __bounds_pointer,
					 __bounds_offset, __bounds_size,
					 __bounds_filename, __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_ptr_plus_int;
#endif
  return __bounds_pnew;
}

#pragma redefine_extname __dummy_bounds_check_ptr_minus_int \
				__bounds_check_ptr_minus_int

extern void *__dummy_bounds_check_ptr_minus_int (__bounds_object *
						 __bounds_obj,
						 void *__bounds_pointer,
						 int __bounds_offset,
						 __bounds_size_t
						 __bounds_size,
						 const char
						 *__bounds_filename,
						 int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_minus_int (__bounds_object * __bounds_obj,
			      void *__bounds_pointer, int __bounds_offset,
			      __bounds_size_t __bounds_size,
			      const char *__bounds_filename,
			      int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pnew;

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer -=
    (int) __bounds_size *__bounds_offset;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_minus_int (__bounds_obj, __bounds_pointer,
					  __bounds_offset, __bounds_size,
					  __bounds_filename, __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_ptr_plus_int;
#endif
  return __bounds_pnew;
}

extern __inline__ void *
__bounds_check_ptr_minus_int_obj (void *__bounds_pointer,
				  int __bounds_offset,
				  __bounds_size_t __bounds_size,
				  const char *__bounds_filename,
				  int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pnew;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pointer);

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer -=
    (int) __bounds_size *__bounds_offset;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_minus_int (__bounds_obj, __bounds_pointer,
					  __bounds_offset, __bounds_size,
					  __bounds_filename, __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_ptr_plus_int;
#endif
  return __bounds_pnew;
}

#pragma redefine_extname __dummy_bounds_check_array_reference \
				__bounds_check_array_reference

extern int
__dummy_bounds_check_array_reference (__bounds_object * __bounds_obj,
				      void *__bounds_pointer,
				      int __bounds_offset,
				      __bounds_size_t __bounds_size,
				      __bounds_size_t __bounds_array_size,
				      const char *__bounds_filename,
				      int __bounds_line);

extern __inline__ int
__bounds_check_array_reference (__bounds_object * __bounds_obj,
				void *__bounds_pointer, int __bounds_offset,
				__bounds_size_t __bounds_size,
				__bounds_size_t __bounds_array_size,
				const char *__bounds_filename,
				int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pnew;
  void *__bounds_ptop;

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer +=
    (int) __bounds_size *__bounds_offset;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       (__bounds_array_size > __bounds_size
	&& (__bounds_offset < 0
	    || ((int) __bounds_size * __bounds_offset) >=
	    (int) __bounds_array_size)) || __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_offset = __dummy_bounds_check_array_reference (__bounds_obj,
							    __bounds_pointer,
							    __bounds_offset,
							    __bounds_size,
							    __bounds_array_size,
							    __bounds_filename,
							    __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_array_reference;
#endif
  return __bounds_offset;
}

extern __inline__ int
__bounds_check_array_reference_obj (void *__bounds_pointer,
				    int __bounds_offset,
				    __bounds_size_t __bounds_size,
				    __bounds_size_t __bounds_array_size,
				    const char *__bounds_filename,
				    int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pnew;
  void *__bounds_ptop;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pointer);

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer +=
    (int) __bounds_size *__bounds_offset;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       (__bounds_array_size > __bounds_size
	&& (__bounds_offset < 0
	    || ((int) __bounds_size * __bounds_offset) >=
	    (int) __bounds_array_size)) || __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_offset = __dummy_bounds_check_array_reference (__bounds_obj,
							    __bounds_pointer,
							    __bounds_offset,
							    __bounds_size,
							    __bounds_array_size,
							    __bounds_filename,
							    __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_array_reference;
#endif
  return __bounds_offset;
}

#pragma redefine_extname __dummy_bounds_check_ptr_diff \
				__bounds_check_ptr_diff

extern int
__dummy_bounds_check_ptr_diff (__bounds_object * __bounds_obj1,
			       __bounds_object * __bounds_obj2,
			       void *__bounds_pointer1,
			       void *__bounds_pointer2,
			       __bounds_size_t __bounds_size,
			       const char *__bounds_filename,
			       int __bounds_line);

extern __inline__ int
__bounds_check_ptr_diff (__bounds_object * __bounds_obj1,
			 __bounds_object * __bounds_obj2,
			 void *__bounds_pointer1, void *__bounds_pointer2,
			 __bounds_size_t __bounds_size,
			 const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_diff (__bounds_obj1, __bounds_obj2,
				     __bounds_pointer1, __bounds_pointer2,
				     __bounds_size, __bounds_filename,
				     __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_diff;
#endif
      __bounds_result = (int) ((char *) __bounds_pointer1 -
			       (char *) __bounds_pointer2) /
	(int) __bounds_size;
    }
  return __bounds_result;
}

extern __inline__ int
__bounds_check_ptr_diff_obj (void *__bounds_pointer1, void *__bounds_pointer2,
			     __bounds_size_t __bounds_size,
			     const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  __bounds_object *__bounds_obj1 = __bounds_find_object (__bounds_pointer1);
  __bounds_object *__bounds_obj2 = __bounds_find_object (__bounds_pointer2);

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_diff (__bounds_obj1, __bounds_obj2,
				     __bounds_pointer1, __bounds_pointer2,
				     __bounds_size, __bounds_filename,
				     __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_diff;
#endif
      __bounds_result = (int) ((char *) __bounds_pointer1 -
			       (char *) __bounds_pointer2) /
	(int) __bounds_size;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_component_reference \
				__bounds_check_component_reference

extern void *__dummy_bounds_check_component_reference (__bounds_object *
						       __bounds_obj,
						       void *__bounds_pointer,
						       int __bounds_offset,
						       int __bounds_size,
						       const char
						       *__bounds_filename,
						       int __bounds_line);

extern __inline__ void *
__bounds_check_component_reference (__bounds_object * __bounds_obj,
				    void *__bounds_pointer,
				    int __bounds_offset, int __bounds_size,
				    const char *__bounds_filename,
				    int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_plow;
  void *__bounds_phigh;

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_offset;
  __bounds_plow = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_phigh = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_plow < __bounds_obj->__bounds_base
       || __bounds_phigh > __bounds_obj->__bounds_extent)
    __bounds_pointer =
      __dummy_bounds_check_component_reference (__bounds_obj,
						__bounds_pointer,
						__bounds_offset,
						__bounds_size,
						__bounds_filename,
						__bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_component_reference;
#endif
  return __bounds_pointer;
}

extern __inline__ void *
__bounds_check_component_reference_obj (void *__bounds_pointer,
					int __bounds_offset,
					int __bounds_size,
					const char *__bounds_filename,
					int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_plow;
  void *__bounds_phigh;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pointer);

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_offset;
  __bounds_plow = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_phigh = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_plow < __bounds_obj->__bounds_base
       || __bounds_phigh > __bounds_obj->__bounds_extent)
    __bounds_pointer =
      __dummy_bounds_check_component_reference (__bounds_obj,
						__bounds_pointer,
						__bounds_offset,
						__bounds_size,
						__bounds_filename,
						__bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_component_reference;
#endif
  return __bounds_pointer;
}

#pragma redefine_extname __dummy_bounds_check_reference \
				__bounds_check_reference

extern void *__dummy_bounds_check_reference (__bounds_object * __bounds_obj,
					     void *__bounds_pointer,
					     __bounds_size_t __bounds_size,
					     const char *__bounds_filename,
					     int __bounds_line);

extern __inline__ void *
__bounds_check_reference (__bounds_object * __bounds_obj,
			  void *__bounds_pointer,
			  __bounds_size_t __bounds_size,
			  const char *__bounds_filename, int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_ptop;

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pointer < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pointer =
      __dummy_bounds_check_reference (__bounds_obj, __bounds_pointer,
				      __bounds_size, __bounds_filename,
				      __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_reference;
#endif
  return __bounds_pointer;
}

extern __inline__ void *
__bounds_check_reference_obj (void *__bounds_pointer,
			      __bounds_size_t __bounds_size,
			      const char *__bounds_filename,
			      int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_ptop;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pointer);

  __bounds_union.__bounds_void_pointer = __bounds_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pointer < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pointer =
      __dummy_bounds_check_reference (__bounds_obj, __bounds_pointer,
				      __bounds_size, __bounds_filename,
				      __bounds_line);
#if __BOUNDS_COLLECT_STATS
  else
    ++__bounds_stats_reference;
#endif
  return __bounds_pointer;
}

#pragma redefine_extname __dummy_bounds_check_ptr_true \
				__bounds_check_ptr_true

extern int
__dummy_bounds_check_ptr_true (void *__bounds_pointer,
			       const char *__bounds_filename,
			       int __bounds_line);

extern __inline__ int
__bounds_check_ptr_true (void *__bounds_pointer,
			 const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_pointer == __BOUNDS_ILLEGAL)
    __bounds_result =
      __dummy_bounds_check_ptr_true (__bounds_pointer, __bounds_filename,
				     __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_true;
#endif
      __bounds_result = __bounds_pointer != __BOUNDS_NULL;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_false \
				__bounds_check_ptr_false

extern int
__dummy_bounds_check_ptr_false (void *__bounds_pointer,
				const char *__bounds_filename,
				int __bounds_line);

extern __inline__ int
__bounds_check_ptr_false (void *__bounds_pointer,
			  const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_pointer == __BOUNDS_ILLEGAL)
    __bounds_result = __dummy_bounds_check_ptr_false (__bounds_pointer,
						      __bounds_filename,
						      __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_false;
#endif
      __bounds_result = __bounds_pointer == __BOUNDS_NULL;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_le_ptr \
				__bounds_check_ptr_le_ptr

extern int
__dummy_bounds_check_ptr_le_ptr (__bounds_object * __bounds_obj1,
				 __bounds_object * __bounds_obj2,
				 void *__bounds_pointer1,
				 void *__bounds_pointer2,
				 const char *__bounds_filename,
				 int __bounds_line);

extern __inline__ int
__bounds_check_ptr_le_ptr (__bounds_object * __bounds_obj1,
			   __bounds_object * __bounds_obj2,
			   void *__bounds_pointer1, void *__bounds_pointer2,
			   const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_le_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_le_ptr;
#endif
      __bounds_result = __bounds_pointer1 <= __bounds_pointer2;
    }
  return __bounds_result;
}

extern __inline__ int
__bounds_check_ptr_le_ptr_obj (void *__bounds_pointer1,
			       void *__bounds_pointer2,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  int __bounds_result;
  __bounds_object *__bounds_obj1 = __bounds_find_object (__bounds_pointer1);
  __bounds_object *__bounds_obj2 = __bounds_find_object (__bounds_pointer2);

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_le_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_le_ptr;
#endif
      __bounds_result = __bounds_pointer1 <= __bounds_pointer2;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_lt_ptr \
				__bounds_check_ptr_lt_ptr

extern int
__dummy_bounds_check_ptr_lt_ptr (__bounds_object * __bounds_obj1,
				 __bounds_object * __bounds_obj2,
				 void *__bounds_pointer1,
				 void *__bounds_pointer2,
				 const char *__bounds_filename,
				 int __bounds_line);

extern __inline__ int
__bounds_check_ptr_lt_ptr (__bounds_object * __bounds_obj1,
			   __bounds_object * __bounds_obj2,
			   void *__bounds_pointer1, void *__bounds_pointer2,
			   const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_lt_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_lt_ptr;
#endif
      __bounds_result = __bounds_pointer1 < __bounds_pointer2;
    }
  return __bounds_result;
}

extern __inline__ int
__bounds_check_ptr_lt_ptr_obj (void *__bounds_pointer1,
			       void *__bounds_pointer2,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  int __bounds_result;
  __bounds_object *__bounds_obj1 = __bounds_find_object (__bounds_pointer1);
  __bounds_object *__bounds_obj2 = __bounds_find_object (__bounds_pointer2);

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_lt_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_lt_ptr;
#endif
      __bounds_result = __bounds_pointer1 < __bounds_pointer2;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_ge_ptr \
				__bounds_check_ptr_ge_ptr

extern int
__dummy_bounds_check_ptr_ge_ptr (__bounds_object * __bounds_obj1,
				 __bounds_object * __bounds_obj2,
				 void *__bounds_pointer1,
				 void *__bounds_pointer2,
				 const char *__bounds_filename,
				 int __bounds_line);

extern __inline__ int
__bounds_check_ptr_ge_ptr (__bounds_object * __bounds_obj1,
			   __bounds_object * __bounds_obj2,
			   void *__bounds_pointer1, void *__bounds_pointer2,
			   const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_ge_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_ge_ptr;
#endif
      __bounds_result = __bounds_pointer1 >= __bounds_pointer2;
    }
  return __bounds_result;
}

extern __inline__ int
__bounds_check_ptr_ge_ptr_obj (void *__bounds_pointer1,
			       void *__bounds_pointer2,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  int __bounds_result;
  __bounds_object *__bounds_obj1 = __bounds_find_object (__bounds_pointer1);
  __bounds_object *__bounds_obj2 = __bounds_find_object (__bounds_pointer2);

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_ge_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_ge_ptr;
#endif
      __bounds_result = __bounds_pointer1 >= __bounds_pointer2;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_gt_ptr \
				__bounds_check_ptr_gt_ptr

extern int
__dummy_bounds_check_ptr_gt_ptr (__bounds_object * __bounds_obj1,
				 __bounds_object * __bounds_obj2,
				 void *__bounds_pointer1,
				 void *__bounds_pointer2,
				 const char *__bounds_filename,
				 int __bounds_line);

extern __inline__ int
__bounds_check_ptr_gt_ptr (__bounds_object * __bounds_obj1,
			   __bounds_object * __bounds_obj2,
			   void *__bounds_pointer1, void *__bounds_pointer2,
			   const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_gt_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_gt_ptr;
#endif
      __bounds_result = __bounds_pointer1 > __bounds_pointer2;
    }
  return __bounds_result;
}

extern __inline__ int
__bounds_check_ptr_gt_ptr_obj (void *__bounds_pointer1,
			       void *__bounds_pointer2,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  int __bounds_result;
  __bounds_object *__bounds_obj1 = __bounds_find_object (__bounds_pointer1);
  __bounds_object *__bounds_obj2 = __bounds_find_object (__bounds_pointer2);

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj1 == __BOUNDS_NULL || __bounds_obj1 != __bounds_obj2)
    __bounds_result =
      __dummy_bounds_check_ptr_gt_ptr (__bounds_obj1, __bounds_obj2,
				       __bounds_pointer1, __bounds_pointer2,
				       __bounds_filename, __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_gt_ptr;
#endif
      __bounds_result = __bounds_pointer1 > __bounds_pointer2;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_ne_ptr \
				__bounds_check_ptr_ne_ptr

extern int
__dummy_bounds_check_ptr_ne_ptr (void *__bounds_pointer1,
				 void *__bounds_pointer2,
				 const char *__bounds_filename,
				 int __bounds_line);

extern __inline__ int
__bounds_check_ptr_ne_ptr (void *__bounds_pointer1, void *__bounds_pointer2,
			   const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_pointer1 == __BOUNDS_ILLEGAL
       || __bounds_pointer2 == __BOUNDS_ILLEGAL)
    __bounds_result = __dummy_bounds_check_ptr_ne_ptr (__bounds_pointer1,
						       __bounds_pointer2,
						       __bounds_filename,
						       __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_ne_ptr;
#endif
      __bounds_result = __bounds_pointer1 != __bounds_pointer2;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_eq_ptr \
				__bounds_check_ptr_eq_ptr

extern int
__dummy_bounds_check_ptr_eq_ptr (void *__bounds_pointer1,
				 void *__bounds_pointer2,
				 const char *__bounds_filename,
				 int __bounds_line);

extern __inline__ int
__bounds_check_ptr_eq_ptr (void *__bounds_pointer1, void *__bounds_pointer2,
			   const char *__bounds_filename, int __bounds_line)
{
  int __bounds_result;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_pointer1 == __BOUNDS_ILLEGAL
       || __bounds_pointer2 == __BOUNDS_ILLEGAL)
    __bounds_result = __dummy_bounds_check_ptr_eq_ptr (__bounds_pointer1,
						       __bounds_pointer2,
						       __bounds_filename,
						       __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_eq_ptr;
#endif
      __bounds_result = __bounds_pointer1 == __bounds_pointer2;
    }
  return __bounds_result;
}

#pragma redefine_extname __dummy_bounds_check_ptr_postinc \
				__bounds_check_ptr_postinc

extern void *__dummy_bounds_check_ptr_postinc (__bounds_object * __bounds_obj,
					       void **__bounds_ptr_to_ptr,
					       int __bounds_inc,
					       const char *__bounds_filename,
					       int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_postinc (__bounds_object * __bounds_obj,
			    void **__bounds_ptr_to_ptr, int __bounds_inc,
			    const char *__bounds_filename, int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postinc (__bounds_obj,
						      __bounds_ptr_to_ptr,
						      __bounds_inc,
						      __bounds_filename,
						      __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

extern __inline__ void *
__bounds_check_ptr_postinc_obj (void **__bounds_ptr_to_ptr,
				int __bounds_inc,
				const char *__bounds_filename,
				int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postinc (__bounds_obj,
						      __bounds_ptr_to_ptr,
						      __bounds_inc,
						      __bounds_filename,
						      __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

#pragma redefine_extname __dummy_bounds_check_ptr_preinc \
				__bounds_check_ptr_preinc

extern void *__dummy_bounds_check_ptr_preinc (__bounds_object * __bounds_obj,
					      void **__bounds_ptr_to_ptr,
					      int __bounds_inc,
					      const char *__bounds_filename,
					      int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_preinc (__bounds_object * __bounds_obj,
			   void **__bounds_ptr_to_ptr, int __bounds_inc,
			   const char *__bounds_filename, int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_preinc (__bounds_obj, __bounds_ptr_to_ptr,
				       __bounds_inc, __bounds_filename,
				       __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_preinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

extern __inline__ void *
__bounds_check_ptr_preinc_obj (void **__bounds_ptr_to_ptr,
			       int __bounds_inc,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_preinc (__bounds_obj, __bounds_ptr_to_ptr,
				       __bounds_inc, __bounds_filename,
				       __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_preinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

#pragma redefine_extname __dummy_bounds_check_ptr_postdec \
				__bounds_check_ptr_postdec

extern void *__dummy_bounds_check_ptr_postdec (__bounds_object * __bounds_obj,
					       void **__bounds_ptr_to_ptr,
					       int __bounds_inc,
					       const char *__bounds_filename,
					       int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_postdec (__bounds_object * __bounds_obj,
			    void **__bounds_ptr_to_ptr, int __bounds_inc,
			    const char *__bounds_filename, int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postdec (__bounds_obj,
						      __bounds_ptr_to_ptr,
						      __bounds_inc,
						      __bounds_filename,
						      __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postdec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

extern __inline__ void *
__bounds_check_ptr_postdec_obj (void **__bounds_ptr_to_ptr,
				int __bounds_inc,
				const char *__bounds_filename,
				int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postdec (__bounds_obj,
						      __bounds_ptr_to_ptr,
						      __bounds_inc,
						      __bounds_filename,
						      __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postdec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

#pragma redefine_extname __dummy_bounds_check_ptr_predec \
				__bounds_check_ptr_predec

extern void *__dummy_bounds_check_ptr_predec (__bounds_object * __bounds_obj,
					      void **__bounds_ptr_to_ptr,
					      int __bounds_inc,
					      const char *__bounds_filename,
					      int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_predec (__bounds_object * __bounds_obj,
			   void **__bounds_ptr_to_ptr, int __bounds_inc,
			   const char *__bounds_filename, int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_predec (__bounds_obj, __bounds_ptr_to_ptr,
				       __bounds_inc, __bounds_filename,
				       __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_predec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

extern __inline__ void *
__bounds_check_ptr_predec_obj (void **__bounds_ptr_to_ptr,
			       int __bounds_inc,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent)
    __bounds_pnew =
      __dummy_bounds_check_ptr_predec (__bounds_obj, __bounds_ptr_to_ptr,
				       __bounds_inc, __bounds_filename,
				       __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_predec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

#pragma redefine_extname __dummy_bounds_check_ptr_postinc_ref \
				__bounds_check_ptr_postinc_ref

extern void *__dummy_bounds_check_ptr_postinc_ref (__bounds_object *
						   __bounds_obj,
						   void **__bounds_ptr_to_ptr,
						   int __bounds_inc,
						   __bounds_size_t
						   __bounds_size,
						   const char
						   *__bounds_filename,
						   int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_postinc_ref (__bounds_object * __bounds_obj,
				void **__bounds_ptr_to_ptr, int __bounds_inc,
				__bounds_size_t __bounds_size,
				const char *__bounds_filename,
				int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_pold < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postinc_ref (__bounds_obj,
							  __bounds_ptr_to_ptr,
							  __bounds_inc,
							  __bounds_size,
							  __bounds_filename,
							  __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

extern __inline__ void *
__bounds_check_ptr_postinc_ref_obj (void **__bounds_ptr_to_ptr,
				    int __bounds_inc,
				    __bounds_size_t __bounds_size,
				    const char *__bounds_filename,
				    int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_pold < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postinc_ref (__bounds_obj,
							  __bounds_ptr_to_ptr,
							  __bounds_inc,
							  __bounds_size,
							  __bounds_filename,
							  __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

#pragma redefine_extname __dummy_bounds_check_ptr_preinc_ref \
				__bounds_check_ptr_preinc_ref

extern void *__dummy_bounds_check_ptr_preinc_ref (__bounds_object *
						  __bounds_obj,
						  void **__bounds_ptr_to_ptr,
						  int __bounds_inc,
						  __bounds_size_t
						  __bounds_size,
						  const char
						  *__bounds_filename,
						  int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_preinc_ref (__bounds_object * __bounds_obj,
			       void **__bounds_ptr_to_ptr, int __bounds_inc,
			       __bounds_size_t __bounds_size,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pnew = __dummy_bounds_check_ptr_preinc_ref (__bounds_obj,
							 __bounds_ptr_to_ptr,
							 __bounds_inc,
							 __bounds_size,
							 __bounds_filename,
							 __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_preinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

extern __inline__ void *
__bounds_check_ptr_preinc_ref_obj (void **__bounds_ptr_to_ptr,
				   int __bounds_inc,
				   __bounds_size_t __bounds_size,
				   const char *__bounds_filename,
				   int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer += __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pnew = __dummy_bounds_check_ptr_preinc_ref (__bounds_obj,
							 __bounds_ptr_to_ptr,
							 __bounds_inc,
							 __bounds_size,
							 __bounds_filename,
							 __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_preinc;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

#pragma redefine_extname __dummy_bounds_check_ptr_postdec_ref \
				__bounds_check_ptr_postdec_ref

extern void *__dummy_bounds_check_ptr_postdec_ref (__bounds_object *
						   __bounds_obj,
						   void **__bounds_ptr_to_ptr,
						   int __bounds_inc,
						   __bounds_size_t
						   __bounds_size,
						   const char
						   *__bounds_filename,
						   int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_postdec_ref (__bounds_object * __bounds_obj,
				void **__bounds_ptr_to_ptr, int __bounds_inc,
				__bounds_size_t __bounds_size,
				const char *__bounds_filename,
				int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_pold < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postdec_ref (__bounds_obj,
							  __bounds_ptr_to_ptr,
							  __bounds_inc,
							  __bounds_size,
							  __bounds_filename,
							  __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postdec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

extern __inline__ void *
__bounds_check_ptr_postdec_ref_obj (void **__bounds_ptr_to_ptr,
				    int __bounds_inc,
				    __bounds_size_t __bounds_size,
				    const char *__bounds_filename,
				    int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_pold < __bounds_obj->__bounds_base
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pold = __dummy_bounds_check_ptr_postdec_ref (__bounds_obj,
							  __bounds_ptr_to_ptr,
							  __bounds_inc,
							  __bounds_size,
							  __bounds_filename,
							  __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_postdec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pold;
}

#pragma redefine_extname __dummy_bounds_check_ptr_predec_ref \
				__bounds_check_ptr_predec_ref

extern void *__dummy_bounds_check_ptr_predec_ref (__bounds_object *
						  __bounds_obj,
						  void **__bounds_ptr_to_ptr,
						  int __bounds_inc,
						  __bounds_size_t
						  __bounds_size,
						  const char
						  *__bounds_filename,
						  int __bounds_line);

extern __inline__ void *
__bounds_check_ptr_predec_ref (__bounds_object * __bounds_obj,
			       void **__bounds_ptr_to_ptr, int __bounds_inc,
			       __bounds_size_t __bounds_size,
			       const char *__bounds_filename,
			       int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pnew = __dummy_bounds_check_ptr_predec_ref (__bounds_obj,
							 __bounds_ptr_to_ptr,
							 __bounds_inc,
							 __bounds_size,
							 __bounds_filename,
							 __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_predec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

extern __inline__ void *
__bounds_check_ptr_predec_ref_obj (void **__bounds_ptr_to_ptr,
				   int __bounds_inc,
				   __bounds_size_t __bounds_size,
				   const char *__bounds_filename,
				   int __bounds_line)
{
  __bounds_union_type __bounds_union;
  void *__bounds_pold = *__bounds_ptr_to_ptr;;
  void *__bounds_pnew;
  void *__bounds_ptop;
  __bounds_object *__bounds_obj = __bounds_find_object (__bounds_pold);

  __bounds_union.__bounds_void_pointer = __bounds_pold;
  __bounds_union.__bounds_char_pointer -= __bounds_inc;
  __bounds_pnew = __bounds_union.__bounds_void_pointer;
  __bounds_union.__bounds_char_pointer += __bounds_size;
  __bounds_ptop = __bounds_union.__bounds_void_pointer;

  if (
#if __BOUNDS_DEBUG_FEATURES
       __bounds_debug_print_calls ||
#endif
       __bounds_obj == __BOUNDS_NULL
       || __bounds_pnew < __bounds_obj->__bounds_base
       || __bounds_pnew > __bounds_obj->__bounds_extent
       || __bounds_ptop > __bounds_obj->__bounds_extent)
    __bounds_pnew = __dummy_bounds_check_ptr_predec_ref (__bounds_obj,
							 __bounds_ptr_to_ptr,
							 __bounds_inc,
							 __bounds_size,
							 __bounds_filename,
							 __bounds_line);
  else
    {
#if __BOUNDS_COLLECT_STATS
      ++__bounds_stats_ptr_predec;
#endif
      *__bounds_ptr_to_ptr = __bounds_pnew;
    }
  return __bounds_pnew;
}

#pragma bounds_checking on

#endif

#else /* not __GNUC__ or not __BOUNDS_CHECKING_ON */

#define UNCHECKED_PTR_DIFF(p,q)			((p) - (q))
#define UNCHECKED_PTR_PLUS_INT(p,i)		((p) + (i))
#define UNCHECKED_PTR_MINUS_INT(p,i)		((p) - (i))

#define BOUNDS_CHECKING_OFF
#define BOUNDS_CHECKING_ON

#define BOUNDS_CHECKING_OFF_DURING(stmt)	do { stmt; } while (0)
#define BOUNDS_CHECKING_OFF_IN_EXPR(expr)	(expr)

#endif /* __GNUC__, __BOUNDS_CHECKING_ON */
#endif /* _bounds_unchecked_h */
